import operator

# If scipy is available, use its spline interpolation function, otherwise, 
# use our implementation (for packaging purposes)
try:
    import scipy.interpolate as interp
except ImportError:
    import PDSim.misc.scipylike as interp
    
import warnings

class Motor(object):
    """
    A class that implements the motor model
    
    This can be either a constant efficiency model
    """
    
    def __init__(self):
        
        self.type = ''
        self.suction_fraction = 0.0
            
    def set_eta(self, eta):
        """
        Set the motor efficiency
        """
        self.eta_motor = eta
        self.type = 'const_eta_motor'
        
    def set_coeffs(self, tau_coeffs, eta_coeffs, omega_coeffs):
        """
        Set the coefficients
        
        Parameters
        ----------
        tau_coeffs: iterable (list or similar)
            Values for the torque [N-m]
        eta_coeffs: iterable (list or similar)
            Values for the efficiency [-] (in the range 0 to 1)
        omega_coeffs: iterable (list or similar)
            Values for the rotational speed [rad/s]
        """
        #Only accept coefficients if they are all the same length
        assert len(tau_coeffs) == len(eta_coeffs) == len(omega_coeffs)
        #Set the local values
        self.tau_coeffs = tau_coeffs
        self.eta_coeffs = eta_coeffs
        self.omega_coeffs = omega_coeffs
        
        #interp1d needs the independent variable to be monotonically increasing
        #Join the lists together 
        zipped = zip(self.tau_coeffs, self.eta_coeffs, self.omega_coeffs)
        #Sort the joined lists by the torque
        zipped = sorted(zipped, key = operator.itemgetter(0))
        #Unzip them back into the original variables
        self.tau_coeffs, self.eta_coeffs, self.omega_coeffs = zip(*zipped)
        self.type = 'motor_map'
        
    def plot_eta(self, ax):
        """
        Plot eta v. torque on the given axis
        """
        pass
    
    def plot_speed(self, ax):
        """
        Plot slip speed v. torque on the given axis
        """
        pass
     
    def invert_map(self, Wdot, kind = 'linear'):
        """
        Invert the map to calculate the speed and the torque based on the power
        the power is given by tau*omega
        
        If a constant efficiency, just return (efficiency,None) tuple
        """
        if not kind == 'linear':
            warnings.warn('invert_map does not take parameter "kind" anymore')
            
        if self.type == 'const_eta_motor':
            return self.eta_motor,None
        else:
            Wdot_coeffs = [tau*omega/1000 for tau,omega in zip(self.tau_coeffs, self.omega_coeffs)]
            #Do the 1D interpolation
            eta_interp = interp.splrep(Wdot_coeffs, self.eta_coeffs, k=2, s=0)
            eta = interp.splev(Wdot, eta_interp)
            omega_interp = interp.splrep(Wdot_coeffs, self.omega_coeffs, k=2, s=0)
            omega = interp.splev(Wdot, omega_interp)
        
            return eta,omega
    
    def apply_map(self, tau, kind = 'linear'):
        """        
        Actually use the motor map to calculate the slip speed and the motor
        efficiency
        
        Parameters
        ----------
        tau : float
            Torque [N-m]
        kind : string, optional
            The kind of interpolation to do (see scipy.interpolate.interp1d) - deprecated
            
        Returns
        -------
        eta : float
            Efficiency [-]
        omega : float
            Rotational speed [rad/s]
        """
        if not kind == 'linear':
            warnings.warn('apply_map does not take parameter "kind" anymore')
            
        eta_interp = interp.splrep(self.tau_coeffs, self.eta_coeffs, k=2, s=0)
        eta = interp.splev(tau, eta_interp)
        omega_interp = interp.splrep(self.tau_coeffs, self.omega_coeffs, k=2, s=0)
        omega = interp.splev(tau, omega_interp)
        
        #Return the values
        return eta, omega



import numpy as np
cimport numpy as np

import cython
cimport cython

from libc.math cimport sqrt,sin,cos,tan,atan2,acos,floor,M_PI as pi,pow

cdef class VdVstruct:
    """
    A struct with values for volume and derivative of volume w.r.t. crank angle
    """

    cdef public double V
    """ Volume [m^3] """

    cdef public double dV
    """ Derivative of volume with respect to crank angle [m^3/radian] """

cdef class HTAnglesClass:
    """
    A struct with angles associated with the calculations needed for the assesment
    of heat transfer in the chambers of the scroll compressor
    """

    cdef public double phi_1_i
    """ Maximum involute angle on the inner involute of the wrap that forms the outer wall of the CV"""
    
    cdef public double phi_2_i
    """ Minimum involute angle on the inner involute of the wrap that forms the outer wall of the CV"""
        
    cdef public double phi_1_o
    """ Maximum involute angle on the outer involute of the wrap that forms the inner wall of the CV"""
        
    cdef public double phi_2_o
    """ Minimum involute angle on the outer involute of the wrap that forms the inner wall of the CV"""

    cdef public double phi_i0
    """ The initial angle on the inner involute of the wrap that forms the outer wall of the CV"""

    cdef public double phi_o0
    """ The initial angle on the outer involute of the wrap that forms the inner wall of the CV"""

cdef enum involute_index:
    INVOLUTE_FI
    INVOLUTE_FO
    INVOLUTE_OI
    INVOLUTE_OO    

cdef class CVInvolute:
    """
    A simple struct to contain the involute angles
    """

    cdef public double phi_max
    """ The maximum involute angle along this involute """

    cdef public double phi_min
    """ The minimum involute angle along this involute """
    
    cdef public double phi_0
    """ The initial involute angle along this involute """

    cdef public double dphi_max_dtheta
    """ The derivative of phi_max w.r.t. crank angle """
    
    cdef public double dphi_min_dtheta
    """ The derivative of phi_min w.r.t. crank angle """

    cdef public involute_index involute
    """ The involute_index of this involute """
    
cdef class CVInvolutes:
    cdef public CVInvolute Inner
    """ The values for the inner involute of this chamber """
    cdef public CVInvolute Outer
    """ The values for the outer involute of this chamber """
    cdef public bint has_line_1 
    """ Boolean for existence of the line #1 """
    cdef public bint has_line_2
    """ Boolean for existence of the line #2 """
    
cdef enum sides:
    UP
    DOWN
    MID
    
cdef enum compressor_CV_indices:
    keyIsa, keyIs1, keyIs2, keyId1, keyId2,
    keyIdd, keyIddd
    keyIc1_1 = 1001, keyIc2_1 = 2001, 
    keyIc1_2 = 1002, keyIc2_2 = 2002,
    keyIc1_3 = 1003, keyIc2_3 = 2003, 
    keyIc1_4 = 1004, keyIc2_4 = 2004,
    keyIc1_5 = 1005, keyIc2_5 = 2005, 
    keyIc1_6 = 1006, keyIc2_6 = 2006,
    keyIc1_7 = 1007, keyIc2_7 = 2007, 
    keyIc1_8 = 1008, keyIc2_8 = 2008,
    keyIc1_9 = 1009, keyIc2_9 = 2009, 
    keyIc1_10 = 1010,  keyIc2_10 = 2010
    
#Take over the geo class with strict typing
cdef class geoVals:
    cdef public double h,ro,rb,t
    cdef public double phi_fi0,phi_fis,phi_fie,phi_fo0,phi_fos,phi_foe, phi_oi0,phi_ois,phi_oie, phi_oo0,phi_oos,phi_ooe
    cdef public double xa_arc1,ya_arc1,ra_arc1,t1_arc1,t2_arc1
    cdef public double xa_arc2,ya_arc2,ra_arc2,t1_arc2,t2_arc2
    cdef public double b_line, t1_line, t2_line, m_line
    cdef public double x0_wall,y0_wall,r_wall
    cdef public double delta_radial, delta_flank
    cdef public double phi_ie_offset, delta_suction_offset
    cdef public double cx_scroll, cy_scroll, V_scroll
    cdef public double Vremove
    cdef public np.ndarray xvec_disc_port, yvec_disc_port
    
    cpdef bint is_symmetric(self)
    cpdef double val_if_symmetric(self, double val) except *
    cpdef copy_inplace(self, geoVals target)

cpdef bytes involute_index_to_key(int index)

cpdef tuple scroll_wrap(geoVals geo)

ctypedef fused double_or_numpy:
    cython.double
    np.ndarray
    
cpdef tuple coords_norm(phi_vec, geoVals geo, double theta, flag = *)
cpdef tuple coords_inv(phi_vec, geoVals geo, double theta, flag = *)
cpdef tuple _coords_inv_np(np.ndarray[np.float_t] phi, geoVals geo, double theta, flag = *)
cpdef tuple _coords_inv_d(double phi, geoVals geo, double theta, flag = *)

cdef _coords_inv_d_int(double phi, geoVals geo, double theta, int flag, double *x, double *y)
cdef coords_inv_dtheta(double phi, geoVals geo, double theta, int inv, double *dx, double *dy)
cdef _dcoords_inv_dphi_int(double phi, geoVals geo,double theta, int flag, double *dxdphi, double *dydphi)

cpdef long get_compressor_CV_index(object key) except *
cpdef long get_compression_chamber_index(long path, long alpha)

cdef inline bint matchpair(long key1, long key2, long target1, long target2):
    return (key1 == target1 and key2 == target2) or (key2 == target1 and key1 == target2)
    
cpdef inline double min2(double a, double b):
    return a if a<b else b

cpdef inline double max2(double a, double b):
    return a if a>b else b
    
cpdef double_or_numpy plus_one(double_or_numpy x)

cpdef double Gr(double phi, geoVals geo, double theta, int inv)
cpdef double dGr_dphi(double phi, geoVals geo, double theta, int inv)
cpdef double dGr_dtheta(double phi, geoVals geo, double theta, int inv)

cpdef VdVstruct VdV(double theta, geoVals geo, CVInvolutes inv)
cpdef dict forces(double theta, geoVals geo, CVInvolutes inv, double A)
cpdef double fxA(double phi, geoVals geo, double theta, involute_index inv)
cpdef double fyA(double phi, geoVals geo, double theta, involute_index inv)
cpdef double fFx_p(double phi, geoVals geo, double theta, involute_index inv)
cpdef double fFy_p(double phi, geoVals geo, double theta, involute_index inv)
cpdef double fMO_p(double phi, geoVals geo, double theta, involute_index inv)

cpdef double involute_heat_transfer(double hc, double hs, double  rb, 
                                  double phi1, double phi2, double phi0, 
                                  double T_scroll, double T_CV, double dT_dphi, 
                                  double phim)



from __future__ import division, print_function

import abc, math
from PDSim.misc.datatypes import arraym
import numpy as np

class AbstractODEIntegrator(object):
    __metaclass__ = abc.ABCMeta
    
    @abc.abstractmethod
    def get_initial_array(self):
        pass
        
    @abc.abstractmethod
    def pre_step_callback(self):
        pass
        
    @abc.abstractmethod
    def post_deriv_callback(self):
        pass
        
    @abc.abstractmethod
    def post_step_callback(self):
        pass
        
    @abc.abstractmethod
    def derivs(self):
        pass
        
    @abc.abstractmethod
    def do_integration(self):
        pass
    
class AbstractSimpleEulerODEIntegrator(AbstractODEIntegrator):
    __metaclass__ = abc.ABCMeta
        
    def do_integration(self,N,tmin,tmax):
        """
        The simple Euler ODE integrator
        
        Parameters
        ----------
        N : integer
            Number of steps taken (not including the initial step).
        tmin : float
            Starting value of the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``]
        tmax : float
            Ending value for the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``] 
            
        Returns
        -------
        abort_flag
            If an abort has been requested (by returning a value other than ``False`` from ``premature_termination``), return value from ``premature_termination``.  ``None`` otherwise
        """
        
        # Step variables
        self.t0 = tmin
        self.h = (tmax-tmin)/(N)
        
        # Get the initial value
        self.xold = self.get_initial_array()
        
        for self.Itheta in range(N):
            
            # Check for termination
            abort = self.premature_termination()
            if abort != False: return abort
            
            # Call the pre-step callback
            self.pre_step_callback()
            
            # Derivatives evaluated at old values of t = t0
            self.f1 = self.derivs(self.t0, self.xold)
            
            # Call post derivative callback after the first derivative evaluation (which might cache values)
            self.post_deriv_callback()
            
            # Calculate the new values
            self.xnew = self.xold + self.h*self.f1
            
            # Everything from this step is finished, now update for the next
            # step coming
            self.t0 += self.h
            self.xold = self.xnew
            
            # Call the post-step callback
            self.post_step_callback()
            
        # Bump up once more
        self.Itheta += 1
        
        # Make sure we end up at the right place
        assert((self.t0 - tmax) < 1e-10)
        
        # No termination was requested
        return False

class AbstractHeunODEIntegrator(AbstractODEIntegrator):
    __metaclass__ = abc.ABCMeta
        
    def do_integration(self, N, tmin, tmax):
        """
        The Heun system of ODE integrator
        
        Parameters
        ----------
        N : integer
            Number of steps taken.  There will be N+1 entries in the state matrices
        tmin : float
            Starting value of the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``]
        tmax : float
            Ending value for the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``] 
            
        Returns
        -------
        abort_flag
            If an abort has been requested (by returning a value other than ``False`` from ``premature_termination``), return value from ``premature_termination``.  ``None`` otherwise
        
        """
        
        # Step variables
        self.t0 = tmin
        self.h = (tmax-tmin)/(N)
        
        # Get the initial value
        self.xold = self.get_initial_array()
        
        for self.Itheta in range(N):
            
            # Check for termination
            abort = self.premature_termination()
            if abort != False: return abort
            
            # Call the pre-step callback
            self.pre_step_callback()
            
            # Step 1: derivatives evaluated at old values
            self.f1 = self.derivs(self.t0, self.xold)
            
            # Call post derivative callback after the first derivative evaluation (which might cache values)
            self.post_deriv_callback()
            
            # Predicted values based on extrapolation using initial derivatives
            self.xtemp = self.xold + self.h*self.f1
            
            # Step 2: Evaluated at predicted step
            self.f2 = self.derivs(self.t0 + self.h, self.xtemp)
            
            # Get corrected values
            self.xnew = self.xold + self.h/2.0*(self.f1 + self.f2)
            
            # Everything from this step is finished, now update for the next
            # step coming
            self.t0 += self.h
            self.xold = self.xnew
            
            # Call the post-step callback
            self.post_step_callback()
        
        # Bump up once more
        self.Itheta += 1
        
        # No termination was requested
        return False
            
class AbstractRK45ODEIntegrator(AbstractODEIntegrator):
    __metaclass__ = abc.ABCMeta
        
    def do_integration(self,
                   tmin=0,
                   tmax=2.0*math.pi,
                   hmin=1e-4,
                   eps_allowed=1e-10,
                   step_relax=0.9,
                   valves_callback = None,
                   UseCashKarp=True,
                   **kwargs):
        """
        
        This function implements an adaptive Runge-Kutta-Feldberg 4th/5th order
        solver for the system of equations
        
        Parameters
        ----------
        hmin : float
            Minimum step size, something like 1e-5 usually is good.  Don't make this too big or you may not be able to get a stable solution
        tmin : float
            Starting value of the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``]
        tmax : float
            Ending value for the independent variable.  ``t`` is in the closed range [``tmin``, ``tmax``]
        eps_allowed : float
            Maximum absolute error of any CV per step allowed.  Don't make this parameter too big or you may not be able to get a stable solution.  Also don't make it too small because then you are going to run into truncation error.
        step_relax : float, optional
            The relaxation factor that is used in the step resizing algorithm.  Should be less than 1.0; you can play with this parameter to improve the adaptive resizing, but should not be necessary.
            
        Returns
        -------
        abort_flag
            If an abort has been requested (by returning a value other than ``False`` from ``premature_termination``), return value from ``premature_termination``.  ``None`` otherwise
        
        Notes
        -----
        
        Mathematically the adaptive solver can be expressed as::
        
            k1=h*dy(xn                                                                   ,t)
            k2=h*dy(xn+1.0/4.0*k1                                                        ,t+1.0/4.0*h)
            k3=h*dy(xn+3.0/32.0*k1+9.0/32.0*k2                                           ,t+3.0/8.0*h)
            k4=h*dy(xn+1932.0/2197.0*k1-7200.0/2197.0*k2+7296.0/2197.0*k3                ,t+12.0/13.0*h)
            k5=h*dy(xn+439.0/216.0*k1-8.0*k2+3680.0/513.0*k3-845.0/4104.0*k4             ,t+h)
            k6=h*dy(xn-8.0/27.0*k1+2.0*k2-3544.0/2565.0*k3+1859.0/4104.0*k4-11.0/40.0*k5 ,t+1.0/2.0*h)

        where the function dy(y,t) returns a vector of the ODE expressions.
        The new value is calculated from::
        
            xnplus=xn+gamma1*k1+gamma2*k2+gamma3*k3+gamma4*k4+gamma5*k5+gamma6*k6

        In the adaptive solver, the errors for a given step can be calculated from::

            error=1.0/360.0*k1-128.0/4275.0*k3-2197.0/75240.0*k4+1.0/50.0*k5+2.0/55.0*k6

        If the maximum absolute error is above allowed error, the step size is decreased and the step is 
        tried again until the error is below tolerance.  If the error is better than required, the step
        size is increased to minimize the number of steps required.
        
        Before the step is run, a callback the ``step_callback`` method of this class is called.  In the ``step_callback`` callback function you can do anything you want, but you must return 
        """
        
        # Get the starting array of variables
        self.xold = self.get_initial_array()
        
        # Start at an index of 0
        self.Itheta = 0
        self.t0 = tmin
        self.tmax = tmax
        self.h = hmin
        
        gamma1=16.0/135.0
        gamma2=0.0
        gamma3=6656.0/12825.0
        gamma4=28561.0/56430.0
        gamma5=-9.0/50.0
        gamma6=2.0/55.0
        
        #t is the independent variable here, where t takes on values in the bounded range [tmin,tmax]
        while (self.t0 < self.tmax - 1e-10):
            
            # Check for termination
            abort = self.premature_termination()
            if abort != False: return abort

            self.stepAccepted = False
            
            while not self.stepAccepted:
                
                # reset the flag
                self.disableAdaptive = False
                
                if self.t0 + self.h > self.tmax:
                    self.disableAdaptive = True
                    self.h = self.tmax - self.t0
            
                # 
                self.pre_step_callback()
                
                # We check stepAccepted again because if the derived class
                # sets the variable stepAccepted, we should not actually do the evaluation
                if not self.stepAccepted:
                    
                    if self.h < hmin and not self.disableAdaptive:
                        # Step is too small, just use the minimum step size
                        self.h = 1.0*hmin
                        self.disableAdaptive = True
                
                    # Step 1: derivatives evaluated at old values
                    self.f1 = self.derivs(self.t0, self.xold)
                    
                    # Call post derivative callback after the first derivative evaluation (which might cache values)
                    self.post_deriv_callback()
                
                    self.xnew1 = self.xold+self.h*(1.0/5.0)*self.f1
                    
                    self.f2 = self.derivs(self.t0+1.0/5.0*self.h, self.xnew1)
                    self.xnew2 = self.xold+self.h*(+3.0/40.0*self.f1+9.0/40.0*self.f2)

                    self.f3 = self.derivs(self.t0+3.0/10.0*self.h, self.xnew2)
                    self.xnew3 = self.xold+self.h*(3.0/10.0*self.f1-9.0/10.0*self.f2+6.0/5.0*self.f3)

                    self.f4 = self.derivs(self.t0+3.0/5.0*self.h, self.xnew3)
                    self.xnew4 = self.xold+self.h*(-11.0/54.0*self.f1+5.0/2.0*self.f2-70/27.0*self.f3+35.0/27.0*self.f4)
                    
                    self.f5 = self.derivs(self.t0+self.h, self.xnew4)
                    self.xnew5 = self.xold+self.h*(1631.0/55296*self.f1+175.0/512.0*self.f2+575.0/13824.0*self.f3+44275.0/110592.0*self.f4+253.0/4096.0*self.f5)
                    
                    self.f6 = self.derivs(self.t0+7/8*self.h, self.xnew5)
                    
                    # Updated values at the next step using 5-th order
                    self.xnew = self.xold + self.h*(37/378*self.f1 + 250/621*self.f3 + 125/594*self.f4 + 512/1771*self.f6)
                    
                    error = self.h*(-277/64512*self.f1+6925/370944*self.f3-6925/202752*self.f4-277.0/14336*self.f5+277/7084*self.f6)

                    max_error = np.sqrt(np.sum(np.power(error, 2)))
                    
                    # If the error is too large, make the step size smaller and try
                    # the step again
                    if (max_error > eps_allowed):
                        if not self.disableAdaptive:
                            # Take a smaller step next time, try again on this step
                            # But only if adaptive mode is on
                            self.h *= step_relax*(eps_allowed/max_error)**(0.3)
                            self.stepAccepted=False
                        else:
                            # Accept the step regardless of whether the error 
                            # is too large or not
                            self.stepAccepted = True
                    else:
                        self.stepAccepted = True
                else:
                    pass
                    # print('accepted')
            
            self.t0 += self.h
            self.Itheta += 1
            self.xold = self.xnew
            
            self.post_step_callback()
            
            # The error is already below the threshold
            if (max_error < eps_allowed and self.disableAdaptive == False and max_error > 0):
                # Take a bigger step next time, since eps_allowed>max_error
                self.h *= step_relax*(eps_allowed/max_error)**(0.2)
               
            if not (self.t0 - tmax) < 1e-3:
                raise AssertionError('(self.t0 - tmax) [' + str(self.t0 - tmax) + '] > 1e-3')
        
        # No termination was requested
        return False
            
if __name__ == '__main__':
    
    class TestIntegrator(object):
        """
        Implements the functions needed to satisfy the ABC requirements
        
        This is the problem from wikipedia, where y' = y, with the explicit solution y = exp(t)
        """
        
        def __init__(self):
            self.x, self.y = [], []
            
        def post_deriv_callback(self): pass
        
        def premature_termination(self): return False
            
        def get_initial_array(self):
            return arraym([1.0])
        
        def pre_step_callback(self): 
            if self.Itheta == 0:
                self.x.append(self.t0)
                self.y.append(self.xold[0])
        
        def post_step_callback(self): 
            self.x.append(self.t0)
            self.y.append(self.xold[0])
        
        def derivs(self, t0, xold):
            return arraym([xold[0]])
            
    class TestEulerIntegrator(TestIntegrator, AbstractSimpleEulerODEIntegrator):
        """ Mixin class using the functions defined in TestIntegrator """
        pass

    class TestHeunIntegrator(TestIntegrator, AbstractHeunODEIntegrator):
        """ Mixin class using the functions defined in TestIntegrator """
        pass
            
    class TestRK45Integrator(TestIntegrator, AbstractRK45ODEIntegrator):
        """ Mixin class using the functions defined in TestIntegrator """
        pass
    
    import matplotlib.pyplot as plt, numpy as np
    
    for N in [4, 11, 21]:
        TEI = TestEulerIntegrator()
        TEI.do_integration(N, 0.0, 4.0)
        plt.plot(TEI.x, TEI.y, 'o-', label = 'Euler: ' + str(N))
        
    for N in [4, 11]:
        THI = TestHeunIntegrator()
        THI.do_integration(N, 0.0, 4.0)
        plt.plot(THI.x, THI.y, '^-', label = 'Heun: ' + str(N))
        
    TRKI = TestRK45Integrator()
    TRKI.do_integration(0.0, 4.0, eps_allowed = 1e-5)
    plt.plot(TRKI.x, TRKI.y, '^-', label = 'RK45')
        
    t = np.linspace(0, 4, 500)
    plt.plot(t, np.exp(t), 'k', lw = 1)
    plt.legend(loc='best')
    plt.show()



from __future__ import division, print_function, absolute_import
from math import pi,cos,sin,sqrt
from PDSim.misc.scipylike import trapz
from PDSim.flow import flow_models
from PDSim.core.core import PDSimCore
from PDSim.misc.datatypes import arraym
from ._recip import _Recip

class Recip(PDSimCore,_Recip):
    """
    Recip is derived from :class:`PDSimCore <PDSim.core.core.PDSimCore>`.
    
    functions V_dV and heat_transfer_callback are provided in _Recip Cython class 
    """
    def __init__(self):
        PDSimCore.__init__(self)
        
    def __getstate__(self):
        """
        A function for preparing class instance for pickling
         
        Combine the dictionaries from the _Recip base class and the Recip
        class when pickling
        """
        py_dict = self.__dict__.copy()
        py_dict.update(_Recip.__cdict__(self))
        return py_dict

    def __setstate__(self, d):
        """
        A function for unpacking class instance for unpickling
        """
        for k,v in d.iteritems():
            setattr(self,k,v)
            
    def V_dV(self, theta):
        "A thin wrapper around Cython code for pickling purposes"
        return _Recip.V_dV(self, theta)
    
    def Vdisp(self):
        """
        Returns displacement of compressor in m\ :math:`^3`\ /revolution
        """
        return 2*self.crank_length*self.A_piston
    
    def V_shell(self, theta):
        
        return self.shell_volume, 0
        
    def TubeCode(self,Tube,**kwargs):
        Tube.Q = flow_models.IsothermalWallTube(Tube.mdot,Tube.State1,Tube.State2,Tube.fixed,Tube.L,Tube.ID,T_wall=self.Tlumps[0])
    
    def Inlet(self, FlowPath, **kwargs):
        try:
            FlowPath.A=0.01**2/4*pi
            mdot=flow_models.IsentropicNozzle(FlowPath.A,FlowPath.State_up,FlowPath.State_down)
            return mdot
        except ZeroDivisionError:
            return 0.0
        
    def PistonLeakage(self, FlowPath, **kwargs):
        try:
            FlowPath.A=self.delta_gap*self.piston_diameter*pi
            mdot=flow_models.IsentropicNozzle(FlowPath.A,FlowPath.State_up,FlowPath.State_down)
            return mdot
        except ZeroDivisionError:
            return 0.0
                
    def Suction(self,FlowPath,**kwargs):
        if FlowPath.key_up=='A':
            ## pressure in compressor higher than the inlet line
            ## valve is closed - no flow
            return 0.0
        else:
            try:
                FlowPath.A=self.suction_valve.A()
                #FlowPath.A=self.A_suction
                mdot=flow_models.IsentropicNozzle(FlowPath.A,FlowPath.State_up,FlowPath.State_down)
                return mdot
            except ZeroDivisionError:
                return 0.0
        
    def Discharge(self,FlowPath,**kwargs):
        if FlowPath.key_down=='A':
            ## pressure in compressor lower than the discharge line
            ## valve is closed - no flow
            return 0.0
        else:
            try:
                FlowPath.A=self.discharge_valve.A()
#                FlowPath.A=self.A_discharge
                mdot=flow_models.IsentropicNozzle(FlowPath.A,FlowPath.State_up,FlowPath.State_down)
                
                return mdot
            except ZeroDivisionError:
                return 0.0

    def heat_transfer_callback(self,theta):
        "A thin wrapper of the Cython code for pickling purposes"
        return _Recip.heat_transfer_callback(self,theta)   
     
#    def heat_transfer_callback(self,theta):
#        T_w  = self.Tlumps[0] #[K]
#        V,dV=self.V_dV(theta) #[m3,m3/radian]
#        D_h = self.piston_diameter #[m]
#        A_ht = pi*self.piston_diameter*(V/self.A_piston) #[m2]
#        
#        Pr = self.CVs['A'].State.Prandtl #[-]
#        rho = self.CVs['A'].State.rho #[kg/m3]
#        k = self.CVs['A'].State.k #[kW/m-K]
#        mu = self.CVs['A'].State.visc #[Pa-s]
#        T = self.CVs['A'].State.T #[K]
#        u = abs(0.5*dV*self.omega/self.A_piston) #[m/s]
#        Re = (rho*u*D_h)/mu #[kg/m3*m/s*m/Pa/s]=[-]
#        
#        h_c = 0.053*(k/D_h)*Pr**(0.6)*Re**(0.8) #[kW/m2/K]
#        Q = h_c*A_ht*(T_w-T)   #Net heat into control volume [kW]
#        if self.CVs.N > 1:
#            return listm([Q,0])
#        else:
#            return listm([Q])
    
    def mechanical_losses(self):
        """
        
        Mathematical Analysis:
        
        .. math::
        
            x = {L_c}\cos \\theta  + \sqrt {L_1^2 - L_c^2{{\\left( {\\sin \\theta } \\right)}^2}}
            
        .. math::
        
            \\dot x =  - {L_c}\\sin \\theta \\dot \\theta  + \\frac{{ - L_c^2\\sin 2\\theta }}{{2\\sqrt {L_1^2 - L_c^2{{\\left( {\\sin \\theta } \\right)}^2}} }}\\dot \\theta
        
        .. math::
            
            \\bar \\dot x =  - \\frac{{\\dot \\theta }}{\\pi }\\int_0^\pi  {\\left[ { - {L_c}\\sin \\theta  + \\frac{{ - L_c^2\\sin 2\\theta }}{{2\\sqrt {L_1^2 - L_c^2{{\\left( {\\sin \\theta } \\right)}^2}} }}} \\right]d\\theta }
        
        .. math::
            
            \\bar \\dot x =  - \\frac{{\\dot \\theta }}{\\pi }\\left[ {{L_c}\\cos \\theta } \\right]_0^\pi  =  - \\frac{{{L_c}\\dot \\theta }}{\\pi }( - 1 - 1) = \\frac{{2{L_c}\\dot \\theta }}{\\pi }
        
        """
        #Oil with viscosity of 10 cSt (=10e-6 m^2/s) and density of 860 kg/m^3
        A_length = pi*(self.piston_diameter)*self.piston_length #[m2]
        u_ave = 2*self.crank_length/pi*self.omega
        F_viscous = (self.mu_oil*A_length*u_ave)/self.delta_gap #[N]
        W_dot_friction = (F_viscous*u_ave)/1000 #[kNm/s]=[kW]
        return W_dot_friction
        
    def ambient_heat_transfer(self):
        """
        The ambient heat transfer for the compressor in kW
        
        Returns a positive value if heat is added to the compressor
        """
        return self.h_shell*self.A_shell*(self.Tamb-self.Tlumps[0]) #[kW]
        
    def lump_energy_balance_callback(self):
        #Mean heat transfer rate from gas to lump using trapezoidal method for numerical integration
        #Note: Qdot_from_gas has opposite sign of heat transfer TO the gas as calculated in the heat_transfer_callback
        #Note: Qdot_from_gas will likely be less than 0 because heat is being removed and delivered to the gas
        self.Qdot_from_gas = -trapz(self.Q[0,0:self.Itheta],self.t[0:self.Itheta])/(self.t[self.Itheta-1]-self.t[0]) #[kW]
        #Mechanical losses are added to the lump
        self.Wdot_mechanical = self.mechanical_losses() #[kW]
        #Heat transfer between the shell and the ambient
        self.Qamb = self.ambient_heat_transfer() #[kW]
        #Note: heat transfer TO the gas in the tubes is given a positive sign
        #      so sign is flipped for the lump
        self.Qtubes = -sum([Tube.Q for Tube in self.Tubes])
        return self.Qdot_from_gas + self.Wdot_mechanical + self.Qamb + self.Qtubes
    
    def pre_solve(self):
        """
        Other calculations that are an indirect result of the inputs
        """
        self.A_piston = pi*(self.piston_diameter)**2/4
        self.V_dead=self.A_piston*self.x_TDC
        self.A_discharge=pi*self.d_discharge**2/4
        self.A_suction=pi*self.d_suction**2/4
        
    #Name gets mangled in the core base class, so un-mangle it
    def _PDSimCore__post_solve(self):
        """
        {\eta _{motor}} = \frac{{{{\dot W}_{shaft}}}}{{{{\dot W}_{shaft}} + {{\dot W}_{motor}}}}
        {\eta _{motor}}\left( {{{\dot W}_{shaft}} + {{\dot W}_{motor}}} \right) = {{\dot W}_{shaft}}
        {{\dot W}_{motor}} = \frac{{{{\dot W}_{shaft}}}}{{{\eta _{motor}}}} - {{\dot W}_{shaft}}
        """
        #Call the base class function
        PDSimCore._PDSimCore__post_solve(self)
        
        #Extra code for the recip
        #Motor losses
        self.Wdot_motor = self.Wdot*(1/self.eta_motor-1)
        
        #Electrical Power
        self.Wdot_electrical = self.Wdot + self.Wdot_motor
        
        #Overall isentropic efficiency
        self.eta_oi = self.Wdot_i/self.Wdot_electrical
        
if __name__=='__main__':
    print("Running this file doesn't do anything")



import cython
cimport cython

import numpy as np
cimport numpy as np

from libcpp.vector cimport vector

from CoolProp.State cimport State

from PDSim.flow.flow import FlowPath
from PDSim.flow.flow cimport FlowPath

from PDSim.misc.datatypes import arraym, empty_arraym
from PDSim.misc.datatypes cimport arraym, empty_arraym
    
from libc.math cimport exp, log, M_PI as pi, M_E as e, sqrt

cpdef double IsentropicNozzle(double A, State State_up, State State_down, int other_output=*)

cdef class FlowFunction(object):
    cpdef double call(self, FlowPath FP) except *
    cpdef resize(self, int Nvalues)
    cpdef public arraym flows
    cpdef public int Nflows
    cpdef public __name__, __strf__
    
cdef class PyFlowFunctionWrapper(FlowFunction):
    cdef public dict kwargs
    cdef public object Function
    
    cpdef double call(self, FlowPath FP) except *
    
cdef class IsentropicNozzleWrapper(FlowFunction):
    cpdef double call(self, FlowPath FP) except *
    
cpdef double FrictionCorrectedIsentropicNozzle(double A, State State_up, State State_down, double delta, int Type, double t = *, double ro = *)

cdef class ValveModel(object):
    cdef public double A_port,A_valve,d_valve,rho_valve,d_port,m_eff,C_D,k_valve,x_stopper
    cdef public bytes key_up, key_down
    cdef public State State_up, State_down
    cdef public arraym xv
    cdef public double x_tr
    
    cpdef set_xv(self, arraym xv)
    cpdef double A(self)
    
    @cython.locals(exists_keys = cython.list, key = cython.bytes)
    cpdef get_States(self, Core)
    
    cdef _pressure_dominant(self,arraym f, double x, double xdot, double rho, double V, double deltap)
    cdef _flux_dominant(self,arraym f, double x, double xdot, double rho, double V)
    cpdef double flow_velocity(self,State State_up, State State_down)
    cpdef arraym derivs(self, Core)
    cpdef dict __cdict__(self)
